module recover16bits_mod
!  VNY 20.05.2004 : This module is aimed at restoring 
!  16 bits (32767 counts) in some of the pixels belonging to
!  bright sources (usually brighter than the magitude B=14).
!  The reason by which these counts are missed in not yet known.
!  VNY 10.11.2004: modifying the module to use it without the source list
!  14.05.2008 VNY protection is introduced into the recoursive 
!                 subroutine in order to limit the number of recoursions
!                 in the unusual test image  
!
   use dal
   use sasdal
   use errorHandling
!   use OmAllocation_mod      ! For memory allocation
!   use OmSourceparam_mod

   implicit none

   private
   integer (int32) :: nRecCalls

   public recover16bits


contains


subroutine recover16bits(xMin, xMax, yMin, yMax, image, binX, binY)
  ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax
  integer(int32), intent(in) :: binX, binY           
  real(single), intent(inout) :: image(xMin:xMax, yMin:yMax)       
  ! Local variables
  integer(int32) :: i, j, k, iSource, i1, j1 
  integer(int32) :: ix, iy, dxy, nHoles, nHolesTotal
  real(kind=single) :: sCounts, maxCounts
  logical :: flagH, flagV
  integer(int32) :: xHole(1:10), yHole(1:10)
!---------------------------------------------------
  real(single), allocatable, dimension(:, :) :: outOSW  
  real(single), dimension(:, :), pointer :: outOSWP  
  character(len=255) :: outSet      
  type(DataSetT) :: outDataSetH           
  type(ArrayT) :: outArrayH   
  integer(int32) :: allocErr   
  integer           :: istat, sPixels, iminus1, iplus1
  integer(int32) :: outDim(2) 
  integer(int32) :: dim(2)
  integer(int32) :: ismoothWindow
  real(single) :: xPos(1:3000), yPos(1:3000), semiA(1:3000), semiB(1:3000), theta(1:3000)
!  real(single) :: xPos(1:50000), yPos(1:50000), semiA(1:50000), semiB(1:50000), theta(1:50000)
  integer(int32) :: nSources, nSourcesMax
  real(single) :: nSigma
  character(len=100) :: comment

  nSourcesMax=3000
!  nSourcesMax=50000
  nHolesTotal=0
  outSet="testimage.fit"
  ismoothWindow=8
  if (binX .gt. 1) ismoothWindow=ismoothWindow/binX
  write(errstr,'(a)') "Recovering possible 16-bits losses in bright sources"
  call message(AppMsg, Verbose)
  dim(1) = lbound(image,1)
  dim(2)= ubound(image,1)
  !write(*,*) " L53 lbound 1 of image: dim(1)=", dim(1), "  dim(2)=", dim(2)
  dim(1) = lbound(image,2)
  dim(2)= ubound(image,2)
  !write(*,*) " L56 lbound 2 of image: dim(1)=", dim(1), "  dim(2)=", dim(2)
  !write(*,*) " L57 RECOVER16 xMin=", xMin, " xMax=", xMax, " yMin=", yMin, " yMax=", yMax
  !write(*,*) " L58 binX=", binX, " binY=", binY, " ismoothWindow=", ismoothWindow
  !write(*,*) "******************************************************************"
  !do j=656,654,-1
  !   write(*,*) (image(i,j), i=260,262)
  !enddo

  ! Creating the output file for checking
  ! Allocate space for the output observation science window.
  allocate (outOSW(xMin: xMax, yMin: yMax), stat = allocErr)
    
  dim(1) = lbound(outOSW,1)
  dim(2)= ubound(outOSW,1)
  !write(*,*) " L91 lbound 1 of outOSW: dim(1)=", dim(1), "  dim(2)=", dim(2)
  dim(1) = lbound(outOSW,2)
  dim(2)= ubound(outOSW,2)
  !write(*,*) " L94 lbound 2 of outOSW: dim(1)=", dim(1), "  dim(2)=", dim(2)

  call CheckAllocation(allocErr, "outOSW")
  !write(*,*) " L87 dimensions of outOSW: dim(1)=", dim(1), "  dim(2)=", dim(2)  

  outOSW=image

  nSigma=3.0 ! to reject faint sources
  call CreateSourceList(outOSW, xMin, xMax, yMin, yMax, ismoothWindow, &
       xPos, yPos, semiA, semiB, theta, nSources, nSourcesMax, nSigma)

  !write(*,*) " L94 after CreateSourceList nSources=", nSources
  outDim(1)=xMax-xMin+1
  outDim(2)=yMax-yMin+1

!*******************************************************************************
! Create a test image
!*******************************************************************************
!  outDataSetH = dataSet(outSet, Create)
!  outArrayH=addArray(outDataSetH, "PRIMARY", real32, outDim, "Counts", "IMAGE")
!
!  outOSWP => real32Array2Data(outArrayH)
!  outOSWP = outOSW 
!  ! Release handles on data.
!  call release(outArrayH) !
!
!  call release(outDataSetH) 
!  write(*,*) "L112 test image has been created"
!*******************************************************************************

  if(allocated(outOSW)) then
     deallocate(outOSW,stat=istat)
     call CheckDeallocation(allocErr, "outOSW")
  end if

  !write(*,*) "******************************************************************"
  !return
  !write(*,*) " Running the blackhole detection algorithm nSources=", nSources 
  k=0
  dxy=16/binX
    do iSource=1, nSources
       ix=int(xPos(iSource))
       iy=int(yPos(iSource))
       ! Compute the number of counts within the source
       sCounts=0.0
       sPixels=0
       maxCounts=0.0
       do i=ix-dxy,ix+dxy

          
          if (i .lt. xMin .or. i .ge. xMax ) cycle
       
          do j=iy-dxy,iy+dxy
             if ( j .ge. yMin .and. j .lt. yMax ) then
                sCounts=sCounts+image(i,j)
                sPixels=sPixels+1
                if (image(i,j) .gt. maxCounts) maxCounts=image(i,j)
             endif
          enddo ! j
       enddo ! i
       if (sPixels .gt. 0) then
          sCounts=sCounts/ real(sPixels)
       else
          sCounts=0.0
       endif
       !if (iSource .eq. 82) write(*,*) " L138 sCounts=", sCounts 
!       if (sCounts .gt. 3000.0) then
       if (sCounts .gt. 500.0) then
          ! search for a hole
          nHoles=0
          xHole=-1
          yHole=-1
          do i=ix-dxy+1,ix+dxy-1

             iminus1 = i - 1     
             iplus1 = i + 1     

             if (iminus1 .lt. xMin .or. iplus1 .ge. xMax) cycle
            

            do j=iy-dxy+1,iy+dxy-1
                flagH=.false.
                flagV=.false.
                if (j-1 .ge. yMin .and. j+1 .lt. yMax) then
                   if (image(iminus1,j) .gt. image(i,j)+3000.0 .and. image(iplus1,j) .gt. image(i,j)+3000.0) flagH=.true.
                   if (image(i,j-1) .gt. image(i,j)+3000.0 .and. image(i,j+1) .gt. image(i,j)+3000.0) flagV=.true.
                endif
                if (flagH .or. flagV) then

                   !write(*,*) " L148 i=", i, " j=", j, " iSource=", iSource, " flagH=", flagH, " flagV=", flagV                   
                   !write(*,*) " ----------------------"
                   !write(*,*) " (i-1,j):", image(i-1,j), " (i,j):", image(i,j), " (i+1,j):", image(i+1,j)
                   !write(*,*) " (i,j-1):", image(i,j-1), " (i,j):", image(i,j), " (i,j+1):", image(i,j+1)
                   !write(*,*) "----------------------"
                   !if (flagH .and. flagV) write(*,*) "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^"!

                   k=k+1
                   i1=i
                   j1=j
                   nHoles=nHoles+1
                   nHolesTotal=nHolesTotal+1
                   if (nHoles .le. 10) then
                      xHole(nHoles)=i
                      yHole(nHoles)=j
                   endif ! nHoles
                   !image(i,j)=image(i,j)+32767.0 ! it could be also 65535
                   image(i,j)=image(i,j)+65535.0
                endif ! flag
             enddo ! j
          enddo ! i
          !if (nHoles .gt. 0) then
          !   write(*,'(i6,i6,a,i5,a,i5, a,f7.1,a, f7.1, a, i3, a, i4, a, i4)') &
          !        k, iSource, " X=", ix, " Y=", iy, " sCounts=", sCounts, &
          !        " maxCounts=", maxCounts, " nHoles=", nHoles, " i1=", i1, " j1=", j1
          !endif ! nHoles
       endif ! sCounts
    enddo  ! iSource
    !write(*,*) "L206 nHolesTotal=", nHolesTotal
    if (nHolesTotal .gt. 0) then
       write(comment, '(a, i5)') "Number of 16-bit recoveries: ", nHolesTotal 
       write (errstr, *) comment
       call warning("allocErr", comment)
    endif
    !write(*,*) " ---------- "
!******************************************************************************

end subroutine recover16bits

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! CreateSourceList
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine CreateSourceList(image, xMin, xMax, yMin, yMax, ismoothWindow, &
     xPos, yPos, semiA, semiB, theta, nSources, nSourcesMax, nSigma)
  ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(inout) :: image(xMin:xMax, yMin:yMax)       
  ! Local variables
  integer(int32) :: i, j, k,l, i1,j1
  integer(int32), intent(in) :: ismoothWindow
  real(single), allocatable, dimension(:, :) :: bkgImage
  real(single), allocatable, dimension(:, :) :: safeImage
  real(single), allocatable, dimension(:, :) :: maskImage
  integer(int32), allocatable, dimension(:, :) :: imaskImage
  integer(int32), allocatable, dimension(:, :) :: imaskImage2
  integer(int32), allocatable, dimension(:, :) :: imaskImage3
  integer(int32), intent(in) :: nSourcesMax
  real(single), intent(out) :: xPos(1:nSourcesMax)
  real(single), intent(out) :: yPos(1:nSourcesMax)
  real(single), intent(out) :: semiA(nSourcesMax)
  real(single), intent(out) :: semiB(nSourcesMax)
  real(single), intent(out) :: theta(nSourcesMax)
  integer(int32), intent(out) :: nSources
  real(single), intent(in) :: nSigma

  real(single) :: sum1,sum2,sum3
  integer(int32) :: allocErr   
  integer           :: istat
  integer(int32) :: dim(2)
  integer(int32) :: idoubleWindow
  real(single), dimension(-ismoothWindow:ismoothWindow,-ismoothWindow:ismoothWindow ) :: template1
  real(single), dimension(-ismoothWindow:ismoothWindow,-2*ismoothWindow-1:2*ismoothWindow+1) :: template2
  real(single) :: umbral
  real(single) :: r11,r12,r21,r22
  integer(int32) :: nPixels, nPixels0
  character(len=1024) :: regionFileName
  real(single) :: sumX, sumY
  real(single) :: sumValue, bkgValue
  real(single) :: x0,y0, x1, y1
  integer(int32) :: flag, radius, reject
  real(single) :: sqrMax, sqrMin, sqrAver
  integer(int32) :: imaskValue
  real(single) :: threshold, maskAver, maskMax

  regionFileName="sources.reg"
  xPos=0.0
  yPos=0.0
  semiA=2.0
  semiB=2.0
  theta=0.0

  template1=0.0
  do i=-ismoothWindow, ismoothWindow
     do j=ismoothWindow-1,ismoothWindow
        k=i*j
        if (k .lt. 0) template1(i,j)=-1.0
        if (k .gt. 0) template1(i,j)=1.0
     enddo
  enddo

  template2=0.0
  do j=ismoothWindow*2-1,ismoothWindow*2+1
     template2(:,j)=1.0
     template2(:,-j)=-1.0
  enddo

  do j=-2,2
     template2(:,j)=1.0
  enddo

  idoubleWindow=ismoothWindow*2+1

  !write(*,*) " L210 xMin=", xMin, " xMax=", xMax, " yMin=", yMin, " yMax=", yMax
  dim(1) = lbound(image,1)
  dim(2)= ubound(image,1)
  !write(*,*) " L213 lbound 1 of image: dim(1)=", dim(1), "  dim(2)=", dim(2)
  dim(1) = lbound(image,2)
  dim(2)= ubound(image,2)
  !write(*,*) " L216 lbound 2 of image: dim(1)=", dim(1), "  dim(2)=", dim(2)

  allocate (bkgImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "bkgImage")  

  allocate (safeImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "safeImage")  

  allocate (maskImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage")  

  allocate (imaskImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage")  

  allocate (imaskImage2(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage2")  

  allocate (imaskImage3(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage3")  


  bkgImage=image
  safeImage=image

  ! shift the image by 3 pixels vertical
  do i=xMin, xMax
     do j=yMin,yMax-3
        image(i,j)=image(i,j)-bkgImage(i,j+3)
     enddo
  enddo
  ! shift the image by 3 pixels horizontal
  bkgImage=image
  do i=xMin, xMax-3
     do j=yMin,yMax
        image(i,j)=image(i,j)-bkgImage(i+3,j)
     enddo
  enddo

  maskImage=0.0
  umbral=10.0
  do i=xMin,xMax
     do j=yMin,yMax!!
        if (i-2 .ge. xMin .and. i+2 .le. xMax) then
           if (j-2 .ge. yMin .and. j+2 .le. yMax) then
              r11=image(i-2,j-2)
              r12=image(i+1,j-2)
              r21=image(i-2,j+1)
              r22=image(i+1,j+1)
              if (r21 .lt. -umbral &
                   .and. r11 .gt. umbral &
                   .and. r22 .gt. umbral &
                   .and. r12 .lt. -umbral) then
                 maskImage(i,j)=maskImage(i,j)+r11-r12-r21+r22
              endif
           endif
        endif!
     enddo
  enddo

  call ComputeBackground(safeImage, maskImage, bkgImage, xMin,xMax,yMin,yMax, maskAver, maskMax)


  image=safeImage-bkgImage
  where (image .lt. 0.0) image=0.0

  sqrMax=0.0
  sqrMin=1.0E25
  sqrAver=0.0
  sum3=0.0
  do i=xMin, xMax
     do j=yMin,yMax
        sum2=sqrt(bkgImage(i,j))
        if (sum2 .gt. sqrMax) sqrMax=sum2
        if (sum2 .lt. sqrMin) sqrMin=sum2
        sqrAver=sqrAver+sum2
        sum3=sum3+1
     enddo
  enddo
  if (sum3 .gt. 1.0) sqrAver=sqrAver/sum3
  !write (*,*) " L309 sqrMax=", sqrMax, " sqrMin=", sqrMin, " sqrAver=" , sqrAver


  maskImage=image-bkgImage*sqrt(bkgImage)

  where (maskImage .lt. 0.0) maskImage=0.0

  ! protect the high s/n sources with a negative boundary
  imaskImage2=0
  where (maskImage .gt. 0.0) imaskImage2=-4
  imaskImage=0
  do i=xMin,xMax
     do j=yMin, yMax
        if (imaskImage2(i,j) .lt. 0) then
           imaskImage(i,j)=-4
           if (i-1 .ge. xMin) imaskImage(i-1,j)=-4
           if (i+1 .le. xMax) imaskImage(i+1,j)=-4
           if (i-2 .ge. xMin) imaskImage(i-2,j)=-4
           if (i+2 .le. xMax) imaskImage(i+2,j)=-4
        endif
     enddo
  enddo
  imaskImage2=imaskImage
  do i=xMin,xMax
     do j=yMin, yMax
        if (imaskImage2(i,j) .lt. 0) then
           imaskImage(i,j)=-4
           if (j-1 .ge. yMin) imaskImage(i,j-1)=-4
           if (j+1 .le. yMax) imaskImage(i,j+1)=-4
           if (j-2 .ge. yMin) imaskImage(i,j-2)=-4
           if (j+2 .le. yMax) imaskImage(i,j+2)=-4
        endif
     enddo
  enddo
  where(maskImage .gt. 0.0) imaskImage=1

!+++++++++++++++++++++++++++++++++++++++++++++++++++++=
! second step
  ! 
  imaskValue=3
  threshold=35.0
  call SourceBoundary(image, bkgImage, imaskImage, xMin, xMax, &
       yMin, yMax, imaskValue, threshold)

!++++++++++++++++++++++++++++++++++++++++++++++++++++
! third step
  !
  imaskValue=2
  threshold=8.0
  call SourceBoundary(image, bkgImage, imaskImage, xMin, xMax, &
       yMin, yMax, imaskValue, threshold)
  !
!++++++++++++++++++++++++++++++++++++++++++++++++++++
! fourth step
  !
  imaskValue=1
  threshold=4.0
  call SourceBoundary(image, bkgImage, imaskImage, xMin, xMax, &
       yMin, yMax, imaskValue, threshold)
  !
  maskImage=real(imaskImage)
!  !*******************************************************
!! gather the information about the sources
  nSources=0
  nRecCalls=0
  do i=xMin,xMax
     do j=yMin,yMax
        if (imaskImage(i,j) .gt. 0) then
           nSources=nSources+1
           if (nSources .gt. nSourcesMax) nSources=nSourcesMax
           imaskImage(i,j)=-nSources
           nPixels=1 ! this will be the number of connected pixels
           sumX=real(i)
           sumY=real(j)
           sumValue=image(i,j)
           bkgValue=bkgImage(i,j)

           ! Check if the pixel is connected to the others (recursive procedure)
           nPixels0=CheckConnectedPixels(image, bkgImage, imaskImage,xMin,xMax,yMin,&
                yMax,i,j,nSources, sumX, sumY, sumValue, bkgValue)
           nPixels=nPixels+nPixels0
           xPos(nSources)=sumX+1.0
           yPos(nSources)=sumY+1.0
           semiA(nSources)=sqrt(real(nPixels))
           semiB(nSources)=semiA(nSources)


           flag=0
           reject=0

           ! Reject the sources with small number of pixels or fainter than a threshold
           if (nSigma .gt. 0.0) then
              if (nPixels .lt. 4 .or. sumValue .lt. nSigma*maskAver) reject=1
           endif

           !write(8,*) " Source=", nSources, " i=", i+1, " j=", j+1, &
           !     " sumX=", sumX, " sumY=", sumY, " nPixels=", nPixels, " v=", &
           !     sumValue, " b=", sqrt(bkgValue), " flag=", flag, " reject=", reject 
           if (reject .gt. 0) then
              nSources=nSources-1
           endif
           nRecCalls=0
        endif ! (maskImage
     enddo ! j=
  enddo ! i=
  !write(*,*) "**************************************************"
  !write(*,*) " Total number of sources=", nSources
  !write(*,*) " maskAver=", maskAver, " maskMax=", maskMax, " 26*sqrt*maskAver=", 26.0*sqrt(maskAver)

  !call WriteSourceRegion(nSources, nSourcesMax, regionFileName, xPos, yPos, semiA, semiB, theta)
  !write(*,*) " Region file written"

  image=maskImage

!  image=image-5.5*sqrt(bkgImage)
!  where (image .lt. 0.0) image=0.0

!*************************************************
  if(allocated(bkgImage)) then
     deallocate(bkgImage,stat=istat)
     call CheckDeallocation(allocErr, "bkgImage")
  end if

  if(allocated(imaskImage)) then
     deallocate(imaskImage,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage")
  end if

  if(allocated(imaskImage2)) then
     deallocate(imaskImage2,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage2")
  end if

  if(allocated(imaskImage3)) then
     deallocate(imaskImage3,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage3")
  end if

  if(allocated(maskImage)) then
     deallocate(maskImage,stat=istat)
     call CheckDeallocation(allocErr, "maskImage")
  end if

  if(allocated(safeImage)) then
     deallocate(safeImage,stat=istat)
     call CheckDeallocation(allocErr, "safeImage")
  end if

end subroutine CreateSourceList

!****************************************************************
! Subroutine to write the source region
!****************************************************************
subroutine WriteSourceRegion(nSources, nSourcesMax, regionFileName, &
     xPos, yPos, semiA, semiB, theta)
  integer(int32), intent(in) :: nSourcesMAx
  real(single), intent(in) :: xPos(1:nSourcesMax), yPos(1:nSourcesMax), &
       semiA(1:nSourcesMax), semiB(1:nSourcesMax), theta(1:nSourcesMax)
  character(len=1024), intent(in) :: regionFileName
  integer(int32), intent(in) :: nSources
  integer(int32), parameter :: ioUnit=2
  integer(int32) :: i

  open(unit = ioUnit, file = regionFileName, status = 'unknown')
  do i = 1, nSources
     write(ioUnit, '('' ELLIPSE('', f7.2,'', '', f7.2, '', '', f6.2, '', '', f6.2, '', '', f6.2, '')'')') &
              xPos(i), yPos(i), semiA(i), semiB(i), theta(i)
    end do
    close(ioUnit)
  !
end subroutine WriteSourceRegion

!****************************************************************
! Subroutine for delimiting boundaries around the sources
!****************************************************************
subroutine SourceBoundary(image, bkgImage, imaskImage, xMin, xMax, yMin, yMax, &
     imaskValue, threshold)

  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: image(xMin:xMax, yMin:yMax) 
  real(single), intent(in) :: bkgImage(xMin:xMax, yMin:yMax) 
  integer(int32), intent(inout) :: imaskImage(xMin:xMax, yMin:yMax)       
  integer(int32), intent(in) :: imaskValue
  real(single), intent(in) :: threshold

  real(single), allocatable, dimension(:, :) :: maskImage
  integer(int32), allocatable, dimension(:, :) :: imaskImage2
  integer(int32), allocatable, dimension(:, :) :: imaskImage3
  integer(int32) :: i,j
  integer(int32) :: allocErr   
  integer           :: istat

  allocate (maskImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage")  
  allocate (imaskImage2(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage2")  
  allocate (imaskImage3(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage3")  

!  maskImage=image-35.0*sqrt(bkgImage)
  maskImage=image-threshold*sqrt(bkgImage)
  !
  where (maskImage .gt. 0.0) imaskImage2=-3
  imaskImage3=0
  do i=xMin,xMax
     do j=yMin, yMax
        if (imaskImage2(i,j) .lt. 0) then
           imaskImage3(i,j)=-imaskValue
           if (i-1 .ge. xMin) imaskImage3(i-1,j)=-imaskValue
           if (i+1 .le. xMax) imaskImage3(i+1,j)=-imaskValue
           if (i-2 .ge. xMin) imaskImage3(i-2,j)=-imaskValue
           if (i+2 .le. xMax) imaskImage3(i+2,j)=-imaskValue
        endif
     enddo
  enddo
  imaskImage2=imaskImage3
  do i=xMin,xMax
     do j=yMin, yMax
        if (imaskImage2(i,j) .lt. 0) then
           imaskImage3(i,j)=-imaskValue
           if (j-1 .ge. yMin) imaskImage3(i,j-1)=-imaskValue
           if (j+1 .le. yMax) imaskImage3(i,j+1)=-imaskValue
           if (j-2 .ge. yMin) imaskImage3(i,j-2)=-imaskValue
           if (j+2 .le. yMax) imaskImage3(i,j+2)=-imaskValue
        endif
     enddo
  enddo
  where (maskImage .gt. 0.0) imaskImage3=1
  imaskImage=imaskImage+imaskImage3

  if(allocated(imaskImage2)) then
     deallocate(imaskImage2,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage2")
  end if
  if(allocated(imaskImage3)) then
     deallocate(imaskImage3,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage3")
  end if
  if(allocated(maskImage)) then
     deallocate(maskImage,stat=istat)
     call CheckDeallocation(allocErr, "maskImage")
  end if
  !
end subroutine SourceBoundary


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! CheckConnectedPixels
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
recursive function  CheckConnectedPixels &
     (image, bkgImage, imaskImage, xMin, xMax, yMin, yMax, i,j, label, &
     sumX, sumY, sumValue, bkgValue) result(nPixels)
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax
  real(single), intent(in) :: image(xMin:xMax, yMin:yMax)
  real(single), intent(in) :: bkgImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(inout) :: imaskImage(xMin:xMax, yMin:yMax)       
  integer(int32), intent(in) :: i,j
  integer(int32), intent(in) :: label
  real(single), intent(inout) :: sumX, sumY, sumValue, bkgValue
  integer(int32) :: k,l
  integer(int32) :: nPixels
  integer(int32) :: nPixels0, nPixels1
  nPixels0=0
  nRecCalls=nRecCalls+1
  do k=i-1,i+1
     do l=j-1,j+1
        if (k .ge. xMin .and. k .le. xMax &
             .and. l .ge. yMin .and. l .le.yMax) then
           if (imaskImage(k,l) .gt. 0 .and. nRecCalls .lt. 2500) then
              nPixels0=nPixels0+1
              imaskImage(k,l)=-label
!              sumX=sumX+real(k)
!              sumY=sumY+real(l)
!              if (k+2 .le. xMax .and. l+2 .le. yMax) then
                 if (image(k,l) .gt. sumValue) then
                    !sumValue=sumValue+image(k+2,l+2)
                    !bkgValue=bkgValue+bkgImage(k+2,l+2)
                    sumValue=image(k,l)
                    bkgValue=bkgImage(k,l)
                    sumX=real(k)
                    sumY=real(l)
                 endif
!              endif
              nPixels1=CheckConnectedPixels(image, bkgImage, imaskImage,xMin,xMax,yMin,yMax,k,&
                   l,label, sumX, sumY, sumValue, bkgValue)
              nPixels0=nPixels0+nPixels1
           endif
        endif
     enddo ! l=
  enddo ! k=
  nPixels=nPixels0
end function CheckConnectedPixels



!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! CheckSourceCounts
! Checking if the maximum corresponds to the central pixel position (for one-pixel 
! sources)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine  CheckSourceCounts &
     (image, bkgImage, xMin, xMax, yMin, yMax, x0, y0, &
     x1, y1, sumValue, bkgValue, radius, flag)
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax
  real(single), intent(in) :: image(xMin:xMax, yMin:yMax)
  real(single), intent(in) :: bkgImage(xMin:xMax, yMin:yMax)
  real(single), intent(in) :: x0,y0 ! original source coordinates
  real(single), intent(out) :: x1,y1 ! corrected source coordinates
  real(single), intent(inout) :: sumValue, bkgValue
  integer(int32), intent(in) :: radius
  integer(int32), intent(out) :: flag
  integer(int32) :: i,j,k,l
  integer(int32) :: imax, jmax
  real(single) :: maxValue
  integer(int32) :: nIter

  flag=0 ! no correction was made 

  nIter=0
  i=int(x0+0.5)
  j=int(y0+0.5)
  imax=i
  jmax=j
101  maxValue=0.0
  do k=i-radius,i+radius
     do l=j-radius,j+radius
        if (k .ge. xMin .and. k .le. xMax &
             .and. l .ge. yMin .and. l .le.yMax) then
           !write(*,*) " k=", k, " l=", l, " image=", image(k,l)
           if (image(k,l) .gt. maxValue) then
              maxValue=image(k,l)
              imax=k
              jmax=l
           endif
        endif
     enddo ! l=
     !write(*,*) "........................."
  enddo ! k=
  sumValue=maxValue
  bkgValue=bkgImage(imax,jmax)
!  write(*,*) " L463 imax=", imax, " jmax=", jmax, " maxValue=", maxValue, " bkgValue=", sqrt(bkgValue)

  x1=real(imax)
  y1=real(jmax)
  if (imax .eq. i .and. jmax .eq. j) return
  nIter=nIter+1
!  if (nIter .gt. 4) return
  flag=nIter
  return
  i=imax
  j=jmax
  goto 101

end subroutine CheckSourceCounts


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! ComputeBackground
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine ComputeBackground(inputImage, maskImage, bkgImage, xMin, xMax, yMin, yMax, &
     maskAver, maskMax)
  ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(in) :: maskImage(xMin:xMax, yMin:yMax)
  real(single), intent(inout) :: bkgImage(xMin:xMax, yMin:yMax)     
  real(single), intent(out) :: maskAver
  real(single), intent(out) :: maskMax
  ! Local variables
  integer(int32) :: i, j, k,l
  integer(int32) :: dim(2)
  real(single) :: sum1,sum2
  integer(int32) :: ismoothWindow, maskWindow
  real(single) :: threshold
  real(single), allocatable, dimension(:) :: vect
  integer(int32) :: vectSize, nElem
  integer(int32) :: allocErr   
  integer           :: istat
  real(single), allocatable, dimension(:, :) :: tmpImage

!---------------------------------------------
 
  ismoothWindow=4
  vectSize=ismoothWindow*16+4

  allocate (tmpImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "tmpImage")  

  allocate (vect(0:vectSize-1), stat = allocErr)
  call CheckAllocation(allocErr, "vect")  

  dim(1) = lbound(inputImage,1)
  dim(2)= ubound(inputImage,1)
  !write(*,*) " L352 lbound 1 of inputImage: dim(1)=", dim(1), "  dim(2)=", dim(2)
  dim(1) = lbound(inputImage,2)
  dim(2)= ubound(inputImage,2)
  !write(*,*) " L355 lbound 2 of inputImage: dim(1)=", dim(1), "  dim(2)=", dim(2)
  !write(*,*) " mod(7,2)=", mod(7,2), " mod(8,2)=", mod(8,2)
  !write(*,*) " modulo(7,2)=", modulo(7,2), " modulo(8,2)=", modulo(8,2)

  !tmpImage=inputImage

! find average and maximal values of the mask
  maskAver=0.0
  maskMax=0.0
  do i=xMin,xMax
     do j=yMin,yMax
        if (maskImage(i,j) .gt. 0.0) then
           maskAver=maskAver+maskImage(i,j)
           sum2=sum2+1.0
           if (maskImage(i,j) .gt. maskMax) maskMax=maskImage(i,j)
        endif
     enddo
  enddo
  if (sum2 .gt. 1.0) maskAver=maskAver/sum2
  threshold=maskAver+0.3*(maskMax-maskAver)
  !write(*,*) " L377 maskAver=", maskAver, " maskMax=", maskMax, " threshold=", threshold, " <-----"

  ! simple smoothing of the input image
  call SimpleImageSmoothing(inputImage, tmpImage, xMin, xMax, yMin, yMax, ismoothWindow)
 
! applying the mask to the smoothed image
  maskWindow=ismoothWindow/2
  do i=xMin,xMax
     do j=yMin,yMax!!
        if (maskImage(i,j) .gt. 0.0) then
           if (maskImage(i,j) .lt. maskAver) then
              do k=i-maskWindow+2,i+maskWindow+2
                 do l=j-maskWindow+2,j+maskWindow+2
                    if (k .ge. xMin .and. k .le. xMax) then
                       if (l .ge. yMin .and. l .le. yMax) then
                          tmpImage(k,l)=0.0
                       endif
                    endif
                 enddo
              enddo
           else
              !if (maskImage(i,j) .lt. maskAver+0.5*maskMax) then
                 do k=i-ismoothWindow+2,i+ismoothWindow+2
                    do l=j-ismoothWindow+2,j+ismoothWindow+2
                       if (k .ge. xMin .and. k .le. xMax) then
                          if (l .ge. yMin .and. l .le. yMax) then
                             tmpImage(k,l)=0.0
                          endif
                       endif
                    enddo
                 enddo
              !endif
           endif
        endif
     enddo
  enddo

! median smoothing of the masked image
  call MedianImageSmoothing(tmpImage, bkgImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call MedianImageSmoothing(bkgImage, tmpImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call MedianImageSmoothing(tmpImage, bkgImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call MedianImageSmoothing(bkgImage, tmpImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call SimpleImageSmoothing(tmpImage, bkgImage, xMin, xMax, yMin, yMax, ismoothWindow)


  if(allocated(vect)) then
     deallocate(vect,stat=istat)
     call CheckDeallocation(allocErr, "vect")
  end if
  if(allocated(tmpImage)) then
     deallocate(tmpImage,stat=istat)
     call CheckDeallocation(allocErr, "tmpImage")
  end if

end subroutine ComputeBackground


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! SimpleImageSmoothing
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine SimpleImageSmoothing(inputImage, outputImage, xMin, xMax, yMin, yMax, ismoothWindow)
    ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(inout) :: outputImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(in) :: ismoothWindow
     
  ! Local variables
  integer(int32) :: i, j, k,l
  real(single) :: sum1,sum2

!---------------------------------------------
  ! simple smoothing of the input image
  do i=xMin,xMax
     do j=yMin,yMax!!
        sum1=0.0
        sum2=0.0
        do k=i-ismoothWindow,i+ismoothWindow
           do l=j-ismoothWindow,j+ismoothWindow
              if (k .ge. xMin .and. k .le. xMax) then
                 if (l .ge. yMin .and. l .le. yMax) then
                    sum1=sum1+inputImage(k,l)
                    sum2=sum2+1.0
                 endif
              endif
           enddo
        enddo
        if (sum2 .gt. 1.0) sum1=sum1/sum2
        outputImage(i,j)=sum1
     enddo
  enddo
end subroutine SimpleImageSmoothing


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! MedianImageSmoothing
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine MedianImageSmoothing(inputImage, outputImage, xMin, xMax, yMin, yMax, ismoothWindow)
    ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(inout) :: outputImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(in) :: ismoothWindow
  real(single), allocatable, dimension(:) :: vect
  integer(int32) :: vectSize, nElem
  integer(int32) :: allocErr   
  integer           :: istat
  integer(int32) :: i, j, k,l
  real(single) :: sum1

!---------------------------------------------
  vectSize=ismoothWindow*16+4

  allocate (vect(0:vectSize-1), stat = allocErr)
  call CheckAllocation(allocErr, "vect")  

! median smoothing of the masked image
  outputImage=0.0
  do i=xMin, xmax
     do j=yMin,yMax
        ! fill in the vector (cross-shaped aperture)
        vect=0.0
        nElem=0
        l=j
        do k=i-ismoothWindow*2,i+ismoothWindow*2
           if (k .ge. xMin .and. k .le. xMax) then
              if (inputImage(k,l) .gt. 0.0) then
                 vect(nElem)=inputImage(k,l)
                 nElem=nElem+1
              endif
           endif
        enddo
        k=i
        do l=j-ismoothWindow*2,j+ismoothWindow*2
           if (l .ge. yMin .and. l .le. yMax) then
              if (inputImage(k,l) .gt. 0.0) then
                 vect(nElem)=inputImage(k,l)
                 nElem=nElem+1
              endif
           endif
        enddo
        if (nElem .gt. 0) then
           sum1=smoothMedian(vect,vectSize,nElem)
           outputImage(i,j)=sum1
        endif
     enddo
  enddo

  if(allocated(vect)) then
     deallocate(vect,stat=istat)
     call CheckDeallocation(allocErr, "vect")
  end if

end subroutine MedianImageSmoothing


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Median smoothing of a vector
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
real function smoothMedian(vect, vectSize, nmax)
  integer(int32), intent(in) :: vectSize
  integer(int32), intent(in) :: nmax ! number of significant elements in the vector, including 0-th
  real(single), intent(in) :: vect(0:vectSize-1)
  real(kind=double) :: a(0:vectSize-1) ! auxiliar vector
  integer :: k,l
  real(single) :: dum
  if (nmax .gt. 2) then
     ! Copy the vector
     a=vect
     ! ordering the vector a
     do k=0,nmax-2
        do l=k+1, nmax-1
           if (a(l) .lt. a(k)) then
              dum=a(k)
              a(k)=a(l)
              a(l)=dum
           endif
        enddo
     enddo
     if (mod(nmax,2) .eq. 0) then
        l=nmax/2
        smoothMedian=(a(l-1)+a(l))/2.0
     else
        l=(nmax-2)/2
        smoothMedian=a(l)
     endif
  else
     smoothMedian=0.0
     do k=0,nmax-1
        smoothMedian=smoothMedian+vect(k)
        dum=dum+1.0
     enddo
     if (dum .gt. 1.0) smoothMedian=smoothMedian/dum
  endif
end function smoothMedian


!***************************************************
! CheckAllocation
!***************************************************
subroutine CheckAllocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a32, a20)') "Could not allocate memory for : ", string 
       write (errstr, *) comment
       call fatal("allocErr") 
    end if
!    noAllocations = noAllocations + 1
end subroutine CheckAllocation
!***************************************************
! CheckDeallocation
!***************************************************
subroutine CheckDeallocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a31, a20)') "Failed to release memory for : ", trim(string) 
       write (errstr, *) comment
       call warning("allocErr", comment) 
    end if
!    noDeallocations = noDeallocations + 1
end subroutine CheckDeallocation

!*********************************************
! CreateRegionFile
!*********************************************
!subroutine CreateRegionFile2(nSources, sourceData, regionFile)
!    integer(int32), intent(in) :: nSources!
!
!    type(sourceParameters), pointer :: sourceData(:)!
!
!    character(*), intent(in) :: regionFile
!    integer(int32) :: i
!    integer(int32), parameter :: ioUnit = 2
! 
!    
!    open(unit = ioUnit, file = regionfile, status = 'unknown')
!    do i = 1, nSources
!        write(ioUnit, '('' ELLIPSE('', f7.2,'', '', f7.2, '', '', f6.2, '', '', f6.2, '', '', f6.2, '')'')') &
!              sourceData(i)%xPos, sourceData(i)%yPos,   & 
!              sourceData(i)%semiA * 3, &    ! / (binx * plateScale), &
!              sourceData(i)%semiB * 3, &   ! / (biny * plateScale), 
!              sourceData(i)%theta
!    end do
!    close(ioUnit)
!    
!end subroutine CreateRegionFile2

end module recover16bits_mod

